// New C++ definitions
// Using templates so no overhead in code size or runtime performance
//

// cleanup:
#ifdef _MMIO_BYTE
#undef _MMIO_BYTE
#endif

#ifdef _MMIO_WORD
#undef _MMIO_WORD
#endif

#ifdef _SFR_IO_ADDR
#undef _SFR_IO_ADDR
#endif

#if defined(_BV)
#undef _BV
#endif

extern "C" {
#include <stdint.h>
#include <stddef.h>
}

template < class ByteType, uint8_t bitNum >
struct MMIO_Reg8Bit
{
public:
    static uint8_t MaskFor(uint8_t bit) { return (uint8_t)((uint8_t)1 << bit); }
    static uint8_t Mask() { return MaskFor(bitNum); }
    uint8_t mask() { return Mask(); }

    template <uint8_t otherBitNum>
        uint8_t operator | (MMIO_Reg8Bit<ByteType, otherBitNum> other)
            { return Mask() | MMIO_Reg8Bit<ByteType, otherBitNum>::Mask(); }
    template <uint8_t otherBitNum>
        uint8_t operator & (MMIO_Reg8Bit<ByteType, otherBitNum> other)
            { return Mask() & MMIO_Reg8Bit<ByteType, otherBitNum>::Mask(); }
    template <uint8_t otherBitNum>
        uint8_t operator ^ (MMIO_Reg8Bit<ByteType, otherBitNum> other)
            { return Mask() ^ MMIO_Reg8Bit<ByteType, otherBitNum>::Mask(); }
    template <uint8_t otherBitNum>
        uint8_t operator ~ () { return ~Mask(); }

    operator uint8_t () { return mask(); }

    void set() volatile { ByteType b; b |= Mask(); }
    void reset() volatile { ByteType b; b &= ~Mask(); }
};

template <uint16_t mem_addr>
struct MMIO_Word
{
public:
    typedef MMIO_Word<mem_addr> MyType;
    typedef volatile uint16_t     IntType;
    typedef IntType *PointerType;
    typedef IntType &  ReferenceType;
    static const uint16_t Addr = mem_addr;
private:
    ReferenceType reg;
public:
    operator PointerType () { return (IntType*)(void*)(Addr); }
    operator ReferenceType () { return * operator PointerType (); }
    MMIO_Word() : reg(*(PointerType)(void*)(Addr)) { }

    void operator  = (const uint16_t &w) { (operator ReferenceType()) = w; }
    void operator |= (const uint16_t &w) { (operator ReferenceType()) |= w; }
    void operator &= (const uint16_t &w) { (operator ReferenceType()) &= w; }
    void operator ^= (const uint16_t &w) { (operator ReferenceType()) ^= w; }
};

template <uint16_t mem_addr>
struct MMIO_Byte
{
public:
    typedef MMIO_Byte<mem_addr> MyType;
    typedef volatile uint8_t     IntType;
    typedef IntType *   PointerType;
    typedef IntType &   ReferenceType;
    static const uint16_t Addr = mem_addr;
private:
    ReferenceType reg;
public:
    uint16_t address() const { return Addr; }
    static uint16_t Address() { return Addr; }
    operator PointerType () { return (IntType*)(void*)(Addr); }
    operator ReferenceType () { return * operator PointerType (); }
    MMIO_Byte() : reg(*(PointerType)(void*)(Addr)) { }

    void operator  = (const uint8_t b) { (operator ReferenceType()) = b; }
    void operator |= (const uint8_t b) { (operator ReferenceType()) |= b; }
    void operator &= (const uint8_t b) { (operator ReferenceType()) &= b; }
    void operator ^= (const uint8_t b) { (operator ReferenceType()) ^= b; }

    void operator = (const int &i) { operator =((uint8_t)(i & 0xFF)); }
    void operator |= (const int &i) { operator |=((uint8_t)(i & 0xFF)); }

    template <uint8_t bitnum>
        void operator = (MMIO_Reg8Bit<MyType, bitnum>& r8b)
        { operator=((uint8_t)(r8b)); }
    template <uint8_t bitnum>
        void operator |= (MMIO_Reg8Bit<MyType, bitnum>& r8b)
        { operator|=((uint8_t)(r8b)); }
};

#define _SFR_IO_ADDR(sfr)   (sfr)::address()

#define _SFR_IO8_DECLARE(name,mem_addr)      MMIO_Byte<mem_addr> name;
#define _SFR_IO16_DECLARE(name,mem_addr)     MMIO_Word<mem_addr> name;
#define _SFR_MEM8_DECLARE(name,mem_addr)      MMIO_Byte<mem_addr> name;
#define _SFR_MEM16_DECLARE(name,mem_addr)     MMIO_Word<mem_addr> name;
#define _MMIO_BYTE_DECLARE(name,mem_addr)    MMIO_Byte<mem_addr> name;
#define _MMIO_WORD_DECLARE(name,mem_addr)    MMIO_Word<mem_addr> name;
// typeof() is a GNU extension
#define _PORTBIT_DECLARE(name,regref,bitnum) MMIO_Reg8Bit<typeof(regref),bitnum> name;
#define _PORTBITG_DECLARE(name,bitnum)       static const uint8_t name = bitnum;

#define _BV(bitname)            (bitname)
