/* Demo of possible port/bit definitions */
#ifndef _AVR_HACKIO_H_
#define _AVR_HACKIO_H_ 1

#define NEW_STYLE_PORTBITS 1

#if !defined(__ASSEMBLER__) && !defined(__cplusplus)
/* Definitions for C code. */

#if 0
/* My proposal is this:
 *
 * We can define port bits with a macro _PORTBIT() (see below for its
 * definitions):
 */
        /* in iom128.h */
        /* 2-wire Serial Interface Control Register */
        #define TWCR        _SFR_MEM8(0x74)
        # /* 2-wire Control Register - TWCR */
        #define    TWINT   _PORTBIT(TWCR,7)
        #define    TWEA    _PORTBIT(TWCR,6)

/* For generic bit names, we can use _PORTBITG(bitNum): */

        /* Timer/Counter Control and Status Register B (generic) */
        #define    ICNC         _PORTBITG(7)
        #define    ICES         _PORTBITG(6)
/*
 * And then we define those macros as appropriate for the environment we're using.
 * End of story. Everybody's happy.
 */
#endif

#if defined(NEW_STYLE_PORTBITS)
/* When you define NEW_STYLE_PORTBITS, you can use either idiom:
 *
 * existing usage still works (for compatibility):
 *  TWCR |= (_BV(TWINT) | _BV(TWEA));
 *
 * or simplified usage:
 *  TWCR |= (TWINT | TWEA); 
 */
#   if defined(_BV)
#       undef _BV
#   endif

#   define _BV(bitname)  (bitname)
#   define _PORTBIT(portname,bitnum)    (1<<(bitnum))
#   define _PORTBITG(bitnum)    (1<<(bitnum))

#else
    /* Existing meanings */
#   define _PORTBIT(portname,bitnum)   (bitnum)
#   define _PORTBITG(bitnum)   (bitnum)

#endif  /* defined(NEW_STYLE_PORTBITS) */

#if 0
    /* Existing definitions (already in sfr_defs.h) */

/* The existing idiom is that register names are lvalues (references) to memory/IO
 * space, and bit names are small numbers representing bit positions.
 * To manipulate individual bits, use the _BV(bitname) macro to convert the bit
 * number to a bit mask.
 *
 * For instance:
 */
        /* in iom128.h */
        /* 2-wire Serial Interface Control Register */
        #define TWCR      _SFR_MEM8(0x74)
        # /* 2-wire Control Register - TWCR */
        #define    TWINT        7
        #define    TWEA         6

        /* in user code */
        TWCR |= (_BV(TWINT) | _BV(TWEA));
        if (TWCR | _BV(TWINT)) { /* ... */ }
        /* or using macros */
        if (bit_is_set(TWCR,TWINT)) { /* ... */ }

/*
 * There is nothing that prevents you from using a bit number with the wrong register.
 */

/* Current definitions for C: */
#define _MMIO_BYTE(mem_addr) (*(volatile uint8_t *)(mem_addr))
#define _MMIO_WORD(mem_addr) (*(volatile uint16_t *)(mem_addr))
#define _BV(bit) (1 << (bit))
#define _VECTOR(N) __vector_ ## N
#define _SFR_MEM8(mem_addr) _MMIO_BYTE(mem_addr)
#define _SFR_MEM16(mem_addr) _MMIO_WORD(mem_addr)
#define _SFR_IO8(io_addr) _MMIO_BYTE((io_addr) + 0x20)
#define _SFR_IO16(io_addr) _MMIO_WORD((io_addr) + 0x20)

#define _SFR_MEM_ADDR(sfr) ((uint16_t) &(sfr))
#define _SFR_IO_ADDR(sfr) (_SFR_MEM_ADDR(sfr) - 0x20)
#define _SFR_IO_REG_P(sfr) (_SFR_MEM_ADDR(sfr) < 0x60)

#define _SFR_ADDR(sfr) _SFR_MEM_ADDR(sfr)
#define bit_is_set(sfr, bit) (_SFR_BYTE(sfr) & _BV(bit))
#define bit_is_clear(sfr, bit) (!(_SFR_BYTE(sfr) & _BV(bit)))
#define loop_until_bit_is_set(sfr, bit) do { } while (bit_is_clear(sfr, bit))
#define loop_until_bit_is_clear(sfr, bit) do { } while (bit_is_set(sfr, bit))

#endif

#elif defined(__cplusplus)

// include C++ template/macro definitions
#include "hackIO++.h"

#elif defined(__ASSEMBLER__)

/* Bitnames are just bit numbers. */
/*  Two defines: _SFR_ASM_COMPAT, __SFR_OFFSET
 *  control meaning of _SFR_ADDR()
 *  A test, _SFR_IO_REG_P(regname) lets you choose between
 *  _SFR_IO_ADDR(regname) (use in/out/cbi/sbi/sbic/sbis)
 *  and _SFR_MEM_ADDR(regname) (use lds/sts)
 */
#       define _PORTBIT(portname,bitnum)   (bitnum)
#       define _PORTBITG(bitnum)   (bitnum)

#endif

#endif 
